@@ -6,6 +6,7 @@
  *
  * Copyright (C) 2005 David Zeuthen, <davidz@redhat.com>
  * Copyright (C) 2005 - 2008 Dan Williams, <dcbw@redhat.com>
+ * Copyright (C) 2005 - 2010 Red Hat, Inc.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -62,6 +63,8 @@
 #define PW_TYPE_ASK	   1
 #define PW_TYPE_UNUSED 2
 
+#define NM_VPNC_LOCAL_PORT_DEFAULT 500
+
 /************** plugin class **************/
 
 static void vpnc_plugin_ui_interface_init (NMVpnPluginUiInterface *iface_class);
@@ -468,24 +471,31 @@ init_plugin_ui (VpncPluginUiWidget *self, NMConnection *connection, GError **err
 		natt_mode = nm_setting_vpn_get_data_item (s_vpn, NM_VPNC_KEY_NAT_TRAVERSAL_MODE);
 
 	gtk_list_store_append (store, &iter);
-	gtk_list_store_set (store, &iter, 0, _("Cisco UDP (default)"), 1, NM_VPNC_NATT_MODE_CISCO, -1);
+	gtk_list_store_set (store, &iter, 0, _("NAT-T when available (default)"), 1, NM_VPNC_NATT_MODE_NATT, -1);
 	if ((active < 0) && natt_mode) {
-		if (!strcmp (natt_mode, NM_VPNC_NATT_MODE_CISCO))
+		if (!strcmp (natt_mode, NM_VPNC_NATT_MODE_NATT))
 			active = 0;
 	}
 
 	gtk_list_store_append (store, &iter);
-	gtk_list_store_set (store, &iter, 0, _("NAT-T"), 1, NM_VPNC_NATT_MODE_NATT, -1);
+	gtk_list_store_set (store, &iter, 0, _("NAT-T always"), 1, NM_VPNC_NATT_MODE_NATT_ALWAYS, -1);
 	if ((active < 0) && natt_mode) {
-		if (!strcmp (natt_mode, NM_VPNC_NATT_MODE_NATT))
+		if (!strcmp (natt_mode, NM_VPNC_NATT_MODE_NATT_ALWAYS))
 			active = 1;
 	}
 
+	gtk_list_store_append (store, &iter);
+	gtk_list_store_set (store, &iter, 0, _("Cisco UDP"), 1, NM_VPNC_NATT_MODE_CISCO, -1);
+	if ((active < 0) && natt_mode) {
+		if (!strcmp (natt_mode, NM_VPNC_NATT_MODE_CISCO))
+			active = 2;
+	}
+
 	gtk_list_store_append (store, &iter);
 	gtk_list_store_set (store, &iter, 0, _("Disabled"), 1, NM_VPNC_NATT_MODE_NONE, -1);
 	if ((active < 0) && natt_mode) {
 		if (!strcmp (natt_mode, NM_VPNC_NATT_MODE_NONE))
-			active = 2;
+			active = 3;
 	}
 
 	widget = glade_xml_get_widget (priv->xml, "natt_combo");
@@ -567,7 +577,7 @@ update_connection (NMVpnPluginUiWidgetInterface *iface,
 {
 	VpncPluginUiWidget *self = VPNC_PLUGIN_UI_WIDGET (iface);
 	VpncPluginUiWidgetPrivate *priv = VPNC_PLUGIN_UI_WIDGET_GET_PRIVATE (self);
-	NMSettingVPN *s_vpn;
+	NMSettingVPN *s_vpn, *s_vpn_orig;
 	GtkWidget *widget;
 	char *str;
 	GtkTreeModel *model;
@@ -660,6 +670,16 @@ update_connection (NMVpnPluginUiWidgetInterface *iface,
 			nm_setting_vpn_add_secret (s_vpn, NM_VPNC_KEY_SECRET, str);
 	}
 
+	/* Local Port is not in GUI (yet?). So when present in the connection,
+	 * copy it from the old VPN setting to the new one to preserve it.
+	 */
+	s_vpn_orig = (NMSettingVPN *) nm_connection_get_setting (connection, NM_TYPE_SETTING_VPN);
+	if (s_vpn_orig) {
+		const char *local_port = nm_setting_vpn_get_data_item (s_vpn_orig, NM_VPNC_KEY_LOCAL_PORT);
+		if (local_port && strlen (local_port))
+			nm_setting_vpn_add_data_item (s_vpn, NM_VPNC_KEY_LOCAL_PORT, local_port);
+	}
+
 	nm_connection_add_setting (connection, NM_SETTING (s_vpn));
 	return TRUE;
 }
@@ -707,7 +727,7 @@ save_secrets (NMVpnPluginUiWidgetInterface *iface,
 	VpncPluginUiWidget *self = VPNC_PLUGIN_UI_WIDGET (iface);
 	VpncPluginUiWidgetPrivate *priv = VPNC_PLUGIN_UI_WIDGET_GET_PRIVATE (self);
 	NMSettingConnection *s_con;
-	const char *str, *id, *uuid;
+	const char *id, *uuid;
 
 	s_con = (NMSettingConnection *) nm_connection_get_setting (connection, NM_TYPE_SETTING_CONNECTION);
 	if (!s_con) {
@@ -977,10 +997,11 @@ import (NMVpnPluginUiInterface *iface, const char *path, GError **error)
 	gboolean bool_value;
 	NMSettingIP4Config *s_ip4;
 	gint val;
+	gboolean found;
 
 	pcf = pcf_file_load (path);
 	if (!pcf) {
-		g_set_error (error, 0, 0, "does not look like a %s VPN connection",
+		g_set_error (error, 0, 0, "does not look like a %s VPN connection (parse failed)",
 		             VPNC_PLUGIN_NAME);
 		return NULL;
 	}
@@ -996,16 +1017,6 @@ import (NMVpnPluginUiInterface *iface, const char *path, GError **error)
 	s_ip4 = NM_SETTING_IP4_CONFIG (nm_setting_ip4_config_new ());
 	nm_connection_add_setting (connection, NM_SETTING (s_ip4));
 
-	/* Connection name */
-	if (pcf_file_lookup_string (pcf, "main", "Description", &buf))
-		g_object_set (s_con, NM_SETTING_CONNECTION_ID, buf, NULL);
-	else {
-		g_set_error (error, 0, 0, "does not look like a %s VPN connection (parse failed)",
-		             VPNC_PLUGIN_NAME);
-		g_object_unref (connection);
-		return NULL;
-	}
-
 	/* Gateway */
 	if (pcf_file_lookup_string (pcf, "main", "Host", &buf))
 		nm_setting_vpn_add_data_item (s_vpn, NM_VPNC_KEY_GATEWAY, buf);
@@ -1028,6 +1039,10 @@ import (NMVpnPluginUiInterface *iface, const char *path, GError **error)
 
 	/* Optional settings */
 
+	/* Connection name */
+	if (pcf_file_lookup_string (pcf, "main", "Description", &buf))
+		g_object_set (s_con, NM_SETTING_CONNECTION_ID, buf, NULL);
+
 	if (pcf_file_lookup_string (pcf, "main", "UserName", &buf))
 		nm_setting_vpn_add_data_item (s_vpn, NM_VPNC_KEY_XAUTH_USER, buf);
 
@@ -1066,22 +1081,43 @@ import (NMVpnPluginUiInterface *iface, const char *path, GError **error)
 			nm_setting_vpn_add_data_item (s_vpn, NM_VPNC_KEY_SINGLE_DES, "yes");
 	}
 
-	/* Default is enabled, only disabled if explicit EnableNat=0 exists */
+	/* Disable all NAT Traversal if explicit EnableNat=0 exists, otherwise
+	 * default to NAT-T which is newer and standardized.  If EnableNat=1, then
+	 * use Cisco-UDP like always; but if the key "X-NM-Use-NAT-T" is set, then
+	 * use NAT-T.  If the key "X-NM-Force-NAT-T" is set then force NAT-T always
+	 * on.  See vpnc documentation for more information on what the different
+	 * NAT modes are.
+	 */
+	nm_setting_vpn_add_data_item (s_vpn,
+	                              NM_VPNC_KEY_NAT_TRAVERSAL_MODE,
+	                              NM_VPNC_NATT_MODE_CISCO);
+
 	if (pcf_file_lookup_bool (pcf, "main", "EnableNat", &bool_value)) {
-		if (!bool_value) {
+		if (bool_value) {
+			gboolean natt = FALSE, force_natt = FALSE;
+
+			if (!pcf_file_lookup_bool (pcf, "main", "X-NM-Use-NAT-T", &natt))
+				natt = FALSE;
+			if (!pcf_file_lookup_bool (pcf, "main", "X-NM-Force-NAT-T", &force_natt))
+				force_natt = FALSE;
+
+			/* force-natt takes precence over plain natt */
+			if (force_natt) {
+				nm_setting_vpn_add_data_item (s_vpn,
+				                              NM_VPNC_KEY_NAT_TRAVERSAL_MODE,
+				                              NM_VPNC_NATT_MODE_NATT_ALWAYS);
+			} else if (natt) {
+				nm_setting_vpn_add_data_item (s_vpn,
+				                              NM_VPNC_KEY_NAT_TRAVERSAL_MODE,
+				                              NM_VPNC_NATT_MODE_NATT);
+			}
+		} else {
 			nm_setting_vpn_add_data_item (s_vpn,
 			                              NM_VPNC_KEY_NAT_TRAVERSAL_MODE,
 			                              NM_VPNC_NATT_MODE_NONE);
 		}
 	}
 
-	/* Default to Cisco UDP */
-	if (!nm_setting_vpn_get_data_item (s_vpn, NM_VPNC_KEY_NAT_TRAVERSAL_MODE)) {
-		nm_setting_vpn_add_data_item (s_vpn,
-		                              NM_VPNC_KEY_NAT_TRAVERSAL_MODE,
-		                              NM_VPNC_NATT_MODE_CISCO);
-	}
-
 	if (pcf_file_lookup_int (pcf, "main", "PeerTimeout", &val)) {
 		if ((val == 0) || ((val >= 10) && (val <= 86400))) {
 			char *tmp = g_strdup_printf ("%d", (gint) val);
@@ -1122,6 +1158,18 @@ import (NMVpnPluginUiInterface *iface, const char *path, GError **error)
 		}
 	}
 
+	/* UseLegacyIKEPort=0 uses dynamic source IKE port instead of 500.
+	 * http://www.cisco.com/en/US/products/sw/secursw/ps2308/products_administration_guide_chapter09186a008015cfdc.html#1192555
+	 * See also: http://support.microsoft.com/kb/928310
+	 */
+	found = pcf_file_lookup_int (pcf, "main", "UseLegacyIKEPort", &val);
+	if (!found || val != 0) {
+		char *tmp;
+		tmp = g_strdup_printf ("%d", (gint) NM_VPNC_LOCAL_PORT_DEFAULT); /* Use default vpnc local port: 500 */
+		nm_setting_vpn_add_data_item (s_vpn, NM_VPNC_KEY_LOCAL_PORT, tmp);
+		g_free (tmp);
+	}
+
 	g_hash_table_destroy (pcf);
 
 	return connection;
@@ -1147,9 +1195,12 @@ export (NMVpnPluginUiInterface *iface,
 	const char *peertimeout = NULL;
 	const char *dhgroup = NULL;
 	GString *routes = NULL;
+	GString *uselegacyikeport = NULL;
 	gboolean success = FALSE;
 	guint32 routes_count = 0;
 	gboolean save_password = FALSE;
+	gboolean use_natt = FALSE;
+	gboolean use_force_natt = FALSE;
 
 	s_con = NM_SETTING_CONNECTION (nm_connection_get_setting (connection, NM_TYPE_SETTING_CONNECTION));
 	s_ip4 = (NMSettingIP4Config *) nm_connection_get_setting (connection, NM_TYPE_SETTING_IP4_CONFIG);
@@ -1191,8 +1242,19 @@ export (NMVpnPluginUiInterface *iface,
 		singledes = TRUE;
 
 	value = nm_setting_vpn_get_data_item (s_vpn, NM_VPNC_KEY_NAT_TRAVERSAL_MODE);
-	if (value && strlen (value) && strcmp (value, NM_VPNC_NATT_MODE_NONE))
-		enablenat = TRUE;
+	if (value && strlen (value)) {
+		if (!strcmp (value, NM_VPNC_NATT_MODE_CISCO)) {
+			enablenat = TRUE;
+			use_natt = FALSE;
+		} else if (!strcmp (value, NM_VPNC_NATT_MODE_NATT)) {
+			enablenat = TRUE;
+			use_natt = TRUE;
+		} else if (!strcmp (value, NM_VPNC_NATT_MODE_NATT_ALWAYS)) {
+			enablenat = TRUE;
+			use_natt = TRUE;
+			use_force_natt = TRUE;
+		}
+	}
 
 	value = nm_setting_vpn_get_data_item (s_vpn, NM_VPNC_KEY_DPD_IDLE_TIMEOUT);
 	if (value && strlen (value))
@@ -1232,6 +1294,11 @@ export (NMVpnPluginUiInterface *iface,
 		routes = NULL;
 	}
 
+	uselegacyikeport = g_string_new ("");
+	value = nm_setting_vpn_get_data_item (s_vpn, NM_VPNC_KEY_LOCAL_PORT);
+	if (!value || !strcmp (value, "0"))
+		g_string_assign (uselegacyikeport, "UseLegacyIKEPort=0\n");
+
 	fprintf (f, 
 		 "[main]\n"
 		 "Description=%s\n"
@@ -1270,6 +1337,9 @@ export (NMVpnPluginUiInterface *iface,
 		 "EnableSplitDNS=1\n"
 		 "SingleDES=%s\n"
 		 "SPPhonebook=\n"
+		 "%s"
+		 "X-NM-Use-NAT-T=%s\n"
+		 "X-NM-Force-NAT-T=%s\n"
 		 "%s\n",
 		 /* Description */   nm_setting_connection_get_id (s_con),
 		 /* Host */          gateway,
@@ -1281,6 +1351,9 @@ export (NMVpnPluginUiInterface *iface,
 		 /* NTDomain */      domain != NULL ? domain : "",
 		 /* PeerTimeout */   peertimeout != NULL ? peertimeout : "0",
 		 /* SingleDES */     singledes ? "1" : "0",
+		 /* UseLegacyIKEPort */ (uselegacyikeport->len) ? uselegacyikeport->str : "",
+		 /* X-NM-Use-NAT-T */ use_natt ? "1" : "0",
+		 /* X-NM-Force-NAT-T */ use_force_natt ? "1" : "0",
 		 /* X-NM-Routes */   (routes && routes->str) ? routes->str : "");
 
 	success = TRUE;
@@ -1288,6 +1361,8 @@ export (NMVpnPluginUiInterface *iface,
 done:
 	if (routes)
 		g_string_free (routes, TRUE);
+	if (uselegacyikeport)
+		g_string_free (uselegacyikeport, TRUE);
 	fclose (f);
 	return success;
 }
@@ -1402,8 +1477,8 @@ vpnc_plugin_ui_interface_init (NMVpnPluginUiInterface *iface_class)
 	/* interface implementation */
 	iface_class->ui_factory = ui_factory;
 	iface_class->get_capabilities = get_capabilities;
-	iface_class->import = import;
-	iface_class->export = export;
+	iface_class->import_from_file = import;
+	iface_class->export_to_file = export;
 	iface_class->get_suggested_name = get_suggested_name;
 	iface_class->delete_connection = delete_connection;
 }
